"""
Reference: https://developers.notion.com/reference/unfurl-attribute-object

A Link Preview is created from an array of unfurl attribute objects.

A Link Preview is a real-time excerpt of authenticated content that unfurls in Notion when an authenticated user shares an enabled link. 
Developers can build Link Preview integrations to customize how links for domains they own look when the links unfurl in a Notion workspace. 
The display of the Link Preview is customizable in terms of content and layout.

TODO: add support for link preview
"""
